/*=============================================================================

  Name: devLonginPad.c

  Abs:  Device support for LCLS Beam Charge 4 channel ADC module
        (Arcturus Coldfire 5282 processor with LTC2208 ADC)
        longin records.
        Derived from $EPICS/base/src/rec/longinRecord.c

  Auth: 16-nov-2005, Dayle Kotturi (dayle): Initial version for LLRF PAD

-----------------------------------------------------------------------------*/

#include "copyright_SLAC.h"     /* SLAC copyright comments */

/*-----------------------------------------------------------------------------

  Mod:  (newest to oldest)
        xx-xxx-2007, Dayle Kotturi: xxxxxxxxxxxxxxxxxxx

=============================================================================*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "alarm.h"
#include "cvtTable.h"
#include "dbDefs.h"
#include "devLib.h"
#include "dbAccess.h"
#include "recSup.h"
#include "callback.h"
#include "devSup.h"
#include "dbScan.h"     /* for SCAN_IO_EVENT */
#include "link.h"
#include "longinRecord.h" /* for longinRecord structure */
#include "epicsExport.h"/* for epicsExportAddress */
#include <debugPrint.h> /* for DEBUGPRINT */ 
#include "drvPad.h"    /* for pad* */

typedef struct liPad_s {
  int routine; 
  int channel; 
  int sample; 
} liPad_t;
 
typedef struct {
  long		number;
  DEVSUPFUN	report;
  DEVSUPFUN	init;
  DEVSUPFUN	init_record;
  DEVSUPFUN	get_ioint_info;
  DEVSUPFUN	read_write;
} LONGDSET;

#ifdef DEBUG_PRINT
int devLonginPadFlag = 4;
#endif

LOCAL long init_longin();
LOCAL long longin_ioinfo();
LOCAL long read_longin();

LONGDSET devLonginPad = {5, NULL, NULL, init_longin, longin_ioinfo, read_longin};
epicsExportAddress(dset, devLonginPad);


LOCAL long init_longin(struct longinRecord *plongin) {
  liPad_t *plip;

  /* Allocate the memory for this records private data structure */
      plip = (liPad_t *) malloc(sizeof(liPad_t)); 
      if (plip == NULL) {
        DEBUGPRINT(DP_ERROR, devLonginPadFlag, 
          ("init_longin: %s: out of memory\n", plongin->name));
        return S_dev_noMemory;
      }

  /* Tell the record about its private data structure */
  plongin->dpvt = plip;

  switch(plongin->inp.type) {
    case(INST_IO): 
      DEBUGPRINT(DP_INFO, devLonginPadFlag, 
        ("init_longin: %s: has parm: %s\n",
        plongin->name, plongin->inp.value.instio.string));
      sscanf(plongin->inp.value.instio.string, "%d %d %d", 
        &(plip->routine), &(plip->channel), &(plip->sample));
      /* NB: "sample" is read in for all longin records, 
         but only used for PAD_BPH. Doing it this way, means the same
         PVs for the sample size and offset can be use regardless of 
         the number of samples actually in use. */
      DEBUGPRINT(DP_INFO, devLonginPadFlag, 
        ("init_longin: %s's dpvt has routine: %d, channel: %d and sample: %d\n", 
        plongin->name, plip->routine, plip->channel, plip->sample));  
      break;

    default:
      DEBUGPRINT(DP_ERROR, devLonginPadFlag, ("init_longin: %s: illegal INP field: %d", plongin->name, (int) plongin->inp.type));
      break;
  }
  return 0;
}

LOCAL long longin_ioinfo( int cmd, struct longinRecord *plongin, IOSCANPVT	*ppvt ) {
  return 0;
}


LOCAL long read_longin(struct longinRecord *plongin) {
  liPad_t *plip;

  /* Set this to facilitate easy access to routine field */ 
  plip = (liPad_t *)plongin->dpvt;

  switch(plip->routine) {
  case 0:
    plongin->val = padGetAverageQ(plip->channel, plip->sample);
    break;

  case 1:
    plongin->val = padGetAverageI(plip->channel, plip->sample);
    break;

  case 2:
    plongin->val = padGetX(plip->channel, plip->sample);
    break;

#ifdef NEVER
  case 4:
    plongin->val = padGetInterceptQ(plip->channel);
    break;

  case 5:
    plongin->val = padGetInterceptI(plip->channel);
    break;
#endif
  case 6:
    plongin->val = padGetIntrCount();
    break;
   
  case 7:
    plongin->val = padGetScanIOCount();
    break;
   
  case 8:
    plongin->val = padGetNumShortsRead(plip->channel, plip->sample);
    break;
    
  case 9:
    plongin->val = padGetSampleSize(plip->channel, plip->sample);
    break;

  case 10:
    plongin->val = padGetSampleOffset(plip->channel, plip->sample);
    break;

  case 11:
    plongin->val = padGetOverflowQCount();
    break;

  case 12:
    plongin->val = padGetOverflowICount();
    break;

  case 13:
    plongin->val = padGetVarianceQ(plip->channel, plip->sample);
    break;

  case 14:
    plongin->val = padGetVarianceI(plip->channel, plip->sample);
    break;

  case 15:
    plongin->val = padGetAverageS(plip->channel, plip->sample);
    break;

  case 16:
    #ifdef USING_TIMER
    plongin->val = padGetTime();
    #else
    plongin->val = 0; 
    #endif
    break;

  default:
    plongin->udf = 1;
    return ERROR;
  }
  DEBUGPRINT(DP_INFO, devLonginPadFlag, ("read_longin: %s: val is: %d", 
    plongin->name, (int) plongin->val));
 
  /* If SCAN is "I/O Intr", use timestamp from Recv task; otherwise leave alone   
  if (plongin->scan == SCAN_IO_EVENT && plongin->tse == -2) { 
    padOverwriting = TRUE;
    plongin->time = pnetGetRmsgTime(0); 
  }*/

  return OK; 
}
