/* The state machine for the PAD handles the transition between running
   and calibrating. The transition to calibrating is triggered by a 
   button press on the "Calibration Control" choice button if "Calib Rqst"
   is selected. Alternatively, you can set the 
   {PRIM}:{LOCA}:{UNIT}:{PAD}_D_CALIB
   PV to 1 or type padSetCalibrationControl(1) at Cexp> 

   As of 9/29/2006, the disabling of interrupts 
   necessary for calibrating is done inside the call to padCalibration.
   The software triggering is done inside padCalReadSignal, which is
   called by padCalibration. The reason for this is because of the fact
   that the PAD non-linearity test also calls padCalReadSignal, repeatedly,
   and so it made sense to put the soft trigger in there. Then, since
   padNonlin also manages the disable/re-enable of interrupts, it was
   deemed more symmetric if padCalibration did this too.  

*/
program sncpad

%%#include "padMacros.h"
%%#include "drvPad.h"

int opState;
int calibCtrl;
int intrCtrl;
int softTrig;
assign opState to "{PRIM}:{LOCA}:{UNIT}:{PAD}_D_STATE";
assign calibCtrl to "{PRIM}:{LOCA}:{UNIT}:{PAD}_D_CALIB";
assign intrCtrl to "{PRIM}:{LOCA}:{UNIT}:{PAD}_INT_CTRL";
assign softTrig to "{PRIM}:{LOCA}:{UNIT}:{PAD}_SOFT_TRG";
 /* notice changes in PV value */
monitor opState;  
monitor calibCtrl; 
monitor intrCtrl; 
monitor softTrig; 

ss pad {
  state init {
    entry {
      printf("sncpad: init\n");
    }
    when(delay(1)) {
      printf("sncpad: operation state set to start-up\n"); 
    } state startup
  }

  state startup {
    entry {
      opState = PAD_STARTING;
      pvPut(opState);
      calibCtrl = PAD_CAL_IDLE;
      pvPut(calibCtrl);
      printf("sncpad: startup\n");
    }
    when(delay(5)) {
      printf("sncpad: startup complete\n");
    } state running
  }

  state running {
    entry {
      opState = PAD_RUNNING;
      pvPut(opState);
      intrCtrl = 1; /* enable interrupts */
      pvPut(intrCtrl);
      printf("sncpad: running. interrupts enabled\n");
    }
    when(delay(1) && (calibCtrl==PAD_CAL_RQST)) {
      printf("sncpad: calibration requested\n");
    } state calibrating

    when(delay(2) && (calibCtrl==PAD_CAL_DONE)) {
      printf("sncpad: running. ignoring calibration done button press\n");
      calibCtrl = PAD_CAL_IDLE;
      pvPut(calibCtrl);
    } state calibrating
  }

  state calibrating {
    entry {
      opState = PAD_CALIBRATING;
      pvPut(opState);

      printf("sncpad: calling calibration subroutine\n");
      /* padCalibration handles disabling interrupts and (indirectly)
         softTrig so no need to do this here. It re-enables interrupts 
         when done. There is more written at the top of this file about this */
      %%padCalibration(); 
    }
    
    /* Sinc calibration is now a one-shot, make it done upon completion of
       call to padCalibration (or on a button press - but calibration done
       button choice could be removed) */
    when(delay(1) || (calibCtrl==PAD_CAL_DONE)) {
      printf("sncpad: calibration done\n");
    } state calibration_done
  }

  state calibration_done {
    /* use this state to reset the channel structs back to
               rf_channels */
    entry {
      opState = PAD_CALIBRATION_DONE;
      pvPut(opState);
    }
    when(delay(1)) { /* give them a chance to "see" the done set;
                       could remove this if need to resume running
                       faster */
      calibCtrl = PAD_CAL_IDLE;
      pvPut(calibCtrl);
      printf("sncpad: resume running\n");
    } state running
  }
}
